/**
* Copyright (c) 2009, Marc 'Foddex' Oude Kotte <foddex@foddex.net>
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of Marc 'Foddex' Oude Kotte nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY Marc 'Foddex' Oude Kotte ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Marc 'Foddex' Oude Kotte BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef FODDEX_JSON_TEMPLATES_H
#define FODDEX_JSON_TEMPLATES_H

#include "json.h"
#include <list>
#include <set>

namespace json {

	//-------------------------------------------------------------------------
	// basic serialize implementation
	//-------------------------------------------------------------------------
	template<typename _T>
	struct serialize {
		static json_block* write( const _T& value ) {
			return value.write_json();
		}
		static void read( _T& value, const json_block* block ) {
			const json_object* object = dynamic_cast<const json_object*>( block );
			if (!object)
				throw json_exception( "failed to cast to json_object" );
			value.read_json( object );
		}
	};

	//-------------------------------------------------------------------------
	// basic type specializations for serialize
	//-------------------------------------------------------------------------
	template<>
	struct serialize<__int64> {
		static json_block* write( __int64 value ) { 
			return new json_integer( value );
		}
		static void read( __int64& value, const json_block* block ) {
			const json_integer* integer = dynamic_cast<const json_integer*>( block );
			if (!integer) {
				if (!dynamic_cast<const json_null*>( block ))
					throw json_exception( "failed to cast to json_integer" );
				else
					value = 0;
			}
			else
				value = integer->get_value();
		}
	};

	template<>
	struct serialize<int> {
		static json_block* write( int value ) { 
			return new json_integer( value );
		}
		static void read( int& value, const json_block* block ) {
			const json_integer* integer = dynamic_cast<const json_integer*>( block );
			if (!integer) {
				if (!dynamic_cast<const json_null*>( block ))
					throw json_exception( "failed to cast to json_integer" );
				else
					value = 0;
			}
			else
				value = static_cast<int>( integer->get_value() );
		}
	};

	template<>
	struct serialize<size_t> {
		static json_block* write( size_t value ) { 
			return new json_integer( value );
		}
		static void read( size_t& value, const json_block* block ) {
			const json_integer* integer = dynamic_cast<const json_integer*>( block );
			if (!integer) {
				if (!dynamic_cast<const json_null*>( block ))
					throw json_exception( "failed to cast to json_integer" );
				else
					value = 0;
			}
			else
				value = static_cast<size_t>( integer->get_value() );
		}
	};

	template<>
	struct serialize<float> {
		static json_block* write( float value ) { 
			return new json_fractional( value ); 
		}
		static void read( float& value, const json_block* block ) {
			const json_fractional* fractional = dynamic_cast<const json_fractional*>( block );
			if (!fractional) {
				if (!dynamic_cast<const json_null*>( block ))
					throw json_exception( "failed to cast to json_fractional" );
				else
					value = 0.0;
			}
			else
				value = static_cast<float>( fractional->get_value() );
		}
	};

	template<>
	struct serialize<string> {
		static json_block* write( const string& value ) { 
			return new json_string( value ); 
		}
		static void read( string& value, const json_block* block ) {
			const json_string* string = dynamic_cast<const json_string*>( block );
			if (!string) {
				if (!dynamic_cast<const json_null*>( block ))
					throw json_exception( "failed to cast to json_string" );
			}
			else
				value = string->get_value();
		}
	};

	template<>
	struct serialize<bool> {
		static json_block* write( bool value ) { 
			return new json_bool( value ); 
		}
		static void read( bool& value, const json_block* block ) {
			const json_bool* boolean = dynamic_cast<const json_bool*>( block );
			if (!boolean) {
				if (!dynamic_cast<const json_null*>( block )) {
					const json_integer* integer = dynamic_cast<const json_integer*>( block );
					if (!integer)
						throw json_exception( "failed to cast to json_bool" );
					else
						value = static_cast<bool>( integer->get_value() );
				} else
					value = false;
			}
			else
				value = boolean->get_value();
		}
	};

	//-------------------------------------------------------------------------
	// generic vector/list/set serializers, creates a json array
	//-------------------------------------------------------------------------
	template<typename _S>
	struct serialize<vector<_S> > {
		static json_block* write( const vector<_S>& value ) {
			json_array* arr = new json_array();
			for (typename vector<_S>::const_iterator it=value.begin(); it!=value.end(); ++it)
				arr->add( serialize<_S>::write( *it ) );
			return arr;
		}
		static void read( vector<_S>& value, const json_block* block ) {
			const json_array* arr = dynamic_cast<const json_array*>( block );
			if (!arr)
				throw json_exception( "failed to cast to json_array" );
			value.resize( arr->size() );
			for (size_t i=0; i<arr->size(); ++i)
				serialize<_S>::read( value[i], arr->get_value( i ) );
		}
	};

	//-------------------------------------------------------------------------
	template<typename _S>
	struct serialize<list<_S> > {
		static json_block* write( const list<_S>& value ) {
			json_array* arr = new json_array();
			for (typename list<_S>::const_iterator it=value.begin(); it!=value.end(); ++it)
				arr->add( serialize<_S>::write( *it ) );
			return arr;
		}
		static void read( list<_S>& value, const json_block* block ) {
			const json_array* arr = dynamic_cast<const json_array*>( block );
			if (!arr)
				throw json_exception( "failed to cast to json_array" );
			for (size_t i=0; i<arr->size(); ++i) {
				_S val;
				serialize<_S>::read( val, arr->get_value( i ) );
				value.push_back( val );
			}
		}
	};

	//-------------------------------------------------------------------------
	template<typename _S>
	struct serialize<set<_S> > {
		static json_block* write( const set<_S>& value ) {
			json_array* arr = new json_array();
			for (typename set<_S>::const_iterator it=value.begin(); it!=value.end(); ++it)
				arr->add( serialize<_S>::write( *it ) );
			return arr;
		}
		static void read( list<_S>& value, const json_block* block ) {
			const json_array* arr = dynamic_cast<const json_array*>( block );
			if (!arr)
				throw json_exception( "failed to cast to json_array" );
			for (size_t i=0; i<arr->size(); ++i) {
				_S val;
				serialize<_S>::read( val, arr->get_value( i ) );
				value.insert( val );
			}
		}
	};

	//-------------------------------------------------------------------------
	// generic map serializer for maps that have a string as their key, 
	// creates a json object
	//-------------------------------------------------------------------------
	template<typename _V>
	struct serialize<map<string,_V> > {
		static json_block* write( const map<string, _V>& value ) {
			json_object* obj = new json_object();
			for (typename map<string, _V>::const_iterator it=value.begin(); it!=value.end(); ++it)
				obj->set( (*it).first, serialize<_V>::write( (*it).second ) );
			return obj;
		}
		static void read( map<string,_V>& value, const json_block* block ) {
			const json_object* obj = dynamic_cast<const json_object*>( block );
			if (!obj)
				throw json_exception( "failed to cast to json_object" );
			for (map<string, json_block*>::const_iterator it=obj->get_members().begin(); it!=obj->get_members().end(); ++it) {
				serialize<_V>::read( value[ (*it).first ], obj->get_value( (*it).first ) );
			}
		}
	};

	//-------------------------------------------------------------------------
	// generic array serializer, requires number of elements to be specified
	// (basically turns everything that points to something into an array)
	//-------------------------------------------------------------------------
	template<typename _S>
	struct serialize<_S*> {
		static json_block* write( const _S* value, size_t num ) {
			json_array* arr = new json_array();
			while (num--)
				arr->add( serialize<_S>::write( value++ ) );
			return arr;
		}
		static void read( _S*& value, const json_block* block ) {
			const json_array* arr = dynamic_cast<const json_array*>( block );
			if (!arr)
				throw json_exception( "failed to cast to json_array" );
			value = new _S[ arr->size() ];
			for (size_t i=0; i<arr->size(); ++i)
				serialize<_S>::read( value[i], arr->get_value( i ) );
		}
	};

	//-------------------------------------------------------------------------
	// serializer
	//-------------------------------------------------------------------------
	struct serializer {
	private:
		json_object*		WriteObject;
		const json_object*	ReadObject;
	public:
		serializer() : WriteObject( new json_object() ), ReadObject(NULL) {}
		serializer( const json_block* block ) : WriteObject(NULL), ReadObject( dynamic_cast<const json_object*>( block ) ) {
			if (!ReadObject)
				throw json_exception( "failed to cast to json_object" );
		}
		serializer( const json_object* object ) : WriteObject(NULL), ReadObject( object ) {}
		~serializer() {
			if (WriteObject)
				delete WriteObject;
		}
		json_object* get_object() {
			json_object* obj = WriteObject;
			WriteObject = NULL;
			return obj;
		}

		template<typename _T>
		void set( const string& name, const _T& value ) {
			WriteObject->set( name, serialize<_T>::write( value ) );
		}

		template<typename _T>
		void get( const string& name, _T& value ) {
			serialize<_T>::read( value, ReadObject->get_value( name ) );
		}
	};

} // json

#endif
