/**
* Copyright (c) 2009, Marc 'Foddex' Oude Kotte <foddex@foddex.net>
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of Marc 'Foddex' Oude Kotte nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY Marc 'Foddex' Oude Kotte ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Marc 'Foddex' Oude Kotte BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef FODDEX_JSON_H
#define FODDEX_JSON_H

#include <stdarg.h>
#include <vector>
#include <map>
#include <string>
using namespace std;

namespace json {

#	ifdef WIN32
#		pragma	warning( disable : 4996 ) 
#		define INT64_FORMAT "I64d"
#		define DOUBLE_FORMAT "I64f"
#	else
#		define strnicmp strncasecmp
#		define _snprintf snprintf
#		define __int64 int64_t
#		define INT64_FORMAT "ld"
#		define DOUBLE_FORMAT "lf"
#	endif

	class json_parser;

	class json_exception {
	private:
		string			Message;
	public:
		json_exception( const string& message ) : Message( message ) {}
		const string& get_message() const { return Message; }
	};

	class json_block {
	public:
		virtual ~json_block() {}
		virtual void write( string& ) const = 0;
		string written() {
			string str;
			write( str );
			return str;
		}
#	ifndef WIN32
	protected:
		static int _snscanf( const char* str, size_t num, const char* format, ... ) {
			char buf[256];
			if (num>sizeof(buf)-1)
				throw json_exception( "num is too big in snscanf" );
			memcpy( buf, str, num );
			buf[num] = 0;
			va_list args;
			va_start(args, format);
			int res = vsscanf( buf, format, args );
			va_end(args); 
			return res;
		}
#	endif
	};

	class json_string : public json_block {
	private:
		string						Value;
	public:
		json_string() {}
		json_string( const string& value ) : Value(value) {}
		json_string( const char* value ) : Value(value) {}
		void write( string& s ) const {
			write_string( s, Value );
		}
		const string& get_value() const { return Value; }
		string& get_value() { return Value; }
		static void write_string( string& output, const string& value ) {
			output += '\"';
			size_t offset = 0;
			while (offset < value.size()) {
				size_t next = value.find_first_of( "\"\\/\b\n\r\t\f", offset );
				if (next > offset) 
					output += (next!=string::npos) ? value.substr( offset, next-offset ) : value.substr( offset );
				if (next != string::npos) {
					switch (value[next]) {
						case '\"':	output += "\\\""; break;
						case '\\':	output += "\\\\"; break;
						case '/':	output += "\\/"; break;
						case '\b':	output += "\\b"; break;
						case '\n':	output += "\\n"; break;
						case '\r':	output += "\\r"; break;
						case '\t':	output += "\\t"; break;
						case '\f':	output += "\\f"; break;
						default:	// really should never happen
							throw json_exception( "unexpected character during string encoding" );
					}
					++next;
				}
				offset = next;
			}
			output += '\"';
		}
		static json_string* read( const string& s, size_t& offset ) {
			if (s[offset++]!='"')
				throw json_exception( "expected opening \" character at start of string" );
			string value;
			while (offset < s.size()) {
				size_t nextQuote = s.find_first_of( "\"", offset );
				size_t nextBackslash = s.find_first_of( '\\', offset );
				if (nextQuote < nextBackslash) {
					// end of string, because a quote is coming before 
					if (nextQuote > offset)
						value.append( &s[offset], nextQuote - offset );
					offset = nextQuote;
					break;
				}
				else if (nextBackslash < nextQuote) {
					// we've got a backslash, append what we have until now, then see what's coming next
					if (nextBackslash > offset)
						value.append( &s[offset], nextBackslash - offset );
					switch (s[nextBackslash+1]) {
						case '\"':	value += '"'; break;
						case '\\':	value += '\\'; break;
						case '/':	value += '/'; break;
						case 'b':	value += '\b'; break;
						case 'n':	value += '\n'; break;
						case 'r':	value += '\r'; break;
						case 't':	value += '\t'; break;
						case 'f':	value += '\f'; break;
						default:	// really should never happen
							throw json_exception( "unexpected escape character during string decoding" );
					}
					offset = nextBackslash+2;
				}
				else 
					throw json_exception( "unterminated string in string" );
			}
			if (s[offset++]!='"')
				throw json_exception( "expected terminating \" character at end of string" );
			return new json_string( value );
		}
	};

	class json_null : public json_block {
	public:
		void write( string& s ) const {
			s += "null";
		}
		static json_null* read( const string& s, size_t& offset ) {
			if (!strnicmp( &s[offset], "null", 4 )) {
				offset += 4;
				return new json_null();
			}
			throw json_exception( "json_null::read expected 'null'" );
		}
	};

	class json_integer : public json_block {
	private:
		__int64					Value;
	public:
		json_integer() : Value(0) {}
		json_integer( __int64 value ) : Value(value) {}
		void write( string& s ) const {
			char buf[64];
			_snprintf( buf, sizeof(buf)-1, "%" INT64_FORMAT , Value );
			s += buf;
		}
		__int64 get_value() const { return Value; }
		static json_integer* read( const string& s, size_t& offset ) {
			bool negative;
			if (s[offset]=='-') {
				negative = true;
				++offset;
			}
			else
				negative = false;
			size_t end = s.find_first_not_of( "0123456789", offset );
			if (end==string::npos)
				end = s.size();
			__int64 value;
			if (_snscanf( s.c_str()+offset, end - offset, "%" INT64_FORMAT, &value )!=1)
				throw json_exception( "failed to parse 64-bits signed integer value" );
			if (negative)
				value *= -1;
			offset += end - offset;
			return new json_integer( value );
		}
	};

	class json_fractional : public json_block {
	private:
		float						Value;
	public:
		json_fractional() : Value(0) {}
		json_fractional( float value ) : Value(value) {}
		void write( string& s ) const {
			char buf[64];
			_snprintf( buf, sizeof(buf)-1, "%f", Value );
			s += buf;
		}
		float get_value() const { return Value; }
		static json_fractional* read( const string& s, size_t& offset ) {
			bool negative;
			if (s[offset]=='-') {
				negative = true;
				++offset;
			}
			else
				negative = false;
			size_t dot = s.find_first_of( '.', offset );
			if (dot==string::npos)
				throw json_exception( "expected . in json_fractional::read" );
			size_t end = s.find_first_not_of( "0123456789", dot+1 );
			float value;
			if (_snscanf( s.c_str()+offset, end - offset, "%f", &value )!=1)
				throw json_exception( "failed to parse float value" );
			if (negative)
				value *= -1;
			offset += end - offset;
			return new json_fractional( value );
		}
	};

	class json_bool : public json_block {
	private:
		bool						Value;
	public:
		json_bool() : Value( false ) {}
		json_bool( bool value ) : Value(value) {}
		void write( string& s ) const {
			s += Value ? "true" : "false";
		}
		bool get_value() const { return Value; }
		static json_bool* read( const string& s, size_t& offset ) {
			if (!strnicmp( &s[offset], "true", 4 )) {
				offset += 4;
				return new json_bool( true ); 
			} else if (!strnicmp( &s[offset], "false", 5 )) {
				offset += 5;
				return new json_bool( false );
			}
			throw json_exception( "json_bool::read expected true or false" );
		}
	};

	class json_object : public json_block {
	private:
		map<string, json_block*>	Members;
	public:
		~json_object() {
			for (map<string, json_block*>::iterator it=Members.begin(); it!=Members.end(); ++it)
				delete (*it).second;
		}
		json_block* get_value( const string& member ) {
			map<string, json_block*>::iterator it = Members.find( member );
			if (it==Members.end())
				throw json_exception( "specified member not present in object" );
			return (*it).second;
		}
		const json_block* get_value( const string& member ) const {
			map<string, json_block*>::const_iterator it = Members.find( member );
			if (it==Members.end())
				throw json_exception( "specified member not present in object" );
			return (*it).second;
		}
		bool has_value( const string& member ) const {
			return Members.find( member )!=Members.end();
		}
		void write( string& s ) const {
			s += '{';
			for (map<string, json_block*>::const_iterator it=Members.begin(); it!=Members.end(); ) {
				json_string::write_string( s, (*it).first );
				s += ':';
				(*it).second->write( s );
				if (++it != Members.end())
					s += ',';
			}
			s += '}';
		}
		json_object* set( const string& member, json_block* value, bool overwrite=false ) {
			if (!value)
				throw json_exception( "value is NULL in json_object::set" );
			if (!Members.insert( map<string, json_block*>::value_type( member, value ) ).second && !overwrite)
				throw json_exception( "member already set in json_object::set" );
			return this;
		}
		size_t size() const { return Members.size(); }
		inline static json_object* read( const string& s, size_t& offset, json_parser* parser );
		const map<string, json_block*>& get_members() const { return Members; }
	};

	class json_array : public json_block {
	private:
		vector<json_block*>			Values;
	public:
		~json_array() {
			for (vector<json_block*>::iterator it=Values.begin(); it!=Values.end(); ++it)
				delete *it;
		}
		json_block* get_value( size_t index ) {
			if (index >= Values.size())
				throw json_exception( "specified member not present in object" );
			return Values[index];
		}
		const json_block* get_value( size_t index ) const {
			if (index >= Values.size())
				throw json_exception( "specified member not present in object" );
			return Values[index];
		}
		void write( string& s ) const {
			s += '[';
			for (vector<json_block*>::const_iterator it=Values.begin(); it!=Values.end(); ) {
				(*it)->write( s );
				if (++it != Values.end())
					s += ',';
			}
			s += ']';
		}
		json_array* add( json_block* value ) {
			if (!value)
				throw json_exception( "value is NULL in json_array::add" );
			Values.push_back( value );
			return this;
		}
		size_t size() const { return Values.size(); }
		inline static json_array* read( const string& s, size_t& offset, json_parser* parser );
	};

	class json_parser {
	private:
		json_block*	Root;
	public:
		json_parser( const string& s ) {
			size_t offset = 0;
			Root = read( s, offset );
		}
		const json_block* get() const { return Root; }
		json_block* get() { return Root; }
		operator json_block*() { return Root; }
		json_block* read( const string& s, size_t& offset ) {
			// skip over whitespace
			skipover_whitespace( s, offset );

			// check what's coming next
			switch (s[offset]) {
				case '[':
					return json_array::read( s, offset, this );
				case '{':
					return json_object::read( s, offset, this );
				case 'n':
					return json_null::read( s, offset );
				case 't':
				case 'f':
					return json_bool::read( s, offset );
				case '\"':
					return json_string::read( s, offset );
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
				case '-':
				{
					size_t nonDigit = s.find_first_not_of( "0123456789", offset );
					if (nonDigit!=string::npos && s[nonDigit]=='.')
						return json_fractional::read( s, offset );
					else
						return json_integer::read( s, offset );
					break;
				}
				default:
					throw json_exception( "badly formatted input string" );
			}
		}
		void skipover_whitespace( const string& s, size_t& offset ) {
			// skip over whitespace
			offset = s.find_first_not_of( "\t\n\r", offset );
			if (offset==string::npos)
				throw json_exception( "unexpected end of input" );
		}
	};

	inline json_array* json_array::read( const string& s, size_t& offset, json_parser* parser ) {
		if (s[offset++]!='[')
			throw json_exception( "expected open accolade in input for array" );
		json_array* arr = new json_array();
		for (;;) {
			parser->skipover_whitespace( s, offset );
			if (s[offset]==']')
				break;
			if (arr->size())
				if (s[offset++]!=',')
					throw json_exception( "expected , in array input" );
			// no need to skip whitespace, parser->read does that for us
			arr->add( parser->read( s, offset ) );
		}
		if (s[offset++]!=']')
			throw json_exception( "expected closing accolade in input for array" );
		return arr;
	}

	inline json_object* json_object::read( const string& s, size_t& offset, json_parser* parser ) {
		if (s[offset++]!='{')
			throw json_exception( "expected open bracket in input for object" );
		json_object* obj = new json_object();
		for (;;) {
			parser->skipover_whitespace( s, offset );
			if (s[offset]=='}')
				break;
			if (obj->size())
				if (s[offset++]!=',')
					throw json_exception( "expected , in object input" );
			json_string* _key = json_string::read( s, offset );
			string key = _key->get_value();
			delete _key;
			parser->skipover_whitespace( s, offset );
			if (s[offset++]!=':')
				throw json_exception( "expected colon in input for object" );
			// no need to skip whitespace, parser->read does that for us
			json_block* value = parser->read( s, offset );
			obj->set( key, value );
		}
		if (s[offset++]!='}')
			throw json_exception( "expected closing bracket in input for object" );
		return obj;
	}

} // json

#endif
